/*-------------------------------------------------------------------
[SCRIPT]  Blocked Process Report Viewer (procdure)
[DATABASE] master
[DESCRIPTION] Journes SQL Server 2013
[SOURCE] Origine http://sqlblockedprocesses.codeplex.com/
[MAJ PAR] DATAFLY - Arian Papillon (prise en charge des XEvents)
[DATEMAJ] 01/12/2013 
-------------------------------------------------------------------*/

USE master
GO

IF OBJECT_ID('sp_blocked_process_report_viewer') IS NULL 
   EXEC ('
   CREATE PROCEDURE dbo.sp_blocked_process_report_viewer 
   AS 
   SELECT ''Replace Me''')
GO

ALTER PROCEDURE dbo.sp_blocked_process_report_viewer
   (
    @Trace NVARCHAR(MAX)
   ,@Type VARCHAR(20) = 'FILE' 
   )
AS 
   SET NOCOUNT ON

-- Validate @Type
   IF ( @Type NOT IN ( 'FILE', 'TABLE', 'XMLFILE', 'XEVENT_RINGBUFFER' ) ) 
      RAISERROR ('The @Type parameter must be ''FILE'', ''TABLE'', ''XMLFILE'', or ''XEVENT_RINGBUFFER''', 11, 1)

   IF ( @Trace LIKE '%.trc'
        AND @Type <> 'FILE'
      ) 
      RAISERROR ('Warning: You specified a .trc trace. You should also specify @Type = ''FILE''', 10, 1)

   IF ( @Trace LIKE '%.xml'
        AND @Type <> 'XMLFILE'
      ) 
      RAISERROR ('Warning: You specified a .xml trace. You should also specify @Type = ''XMLFILE''', 10, 1)
	

   CREATE TABLE #ReportsXML
      (
       monitorloop NVARCHAR(100) NOT NULL
      ,endTime DATETIME NULL
      ,blocking_spid INT NOT NULL
      ,blocking_ecid INT NOT NULL
      ,blocked_spid INT NOT NULL
      ,blocked_ecid INT NOT NULL
      ,blocked_hierarchy_string AS CAST(blocked_spid AS VARCHAR(20)) + '.' + CAST(blocked_ecid AS VARCHAR(20)) + '/'
      ,blocking_hierarchy_string AS CAST(blocking_spid AS VARCHAR(20)) + '.' + CAST(blocking_ecid AS VARCHAR(20)) + '/'
      ,bpReportXml XML NOT NULL
      ,PRIMARY KEY CLUSTERED ( monitorloop, blocked_spid, blocked_ecid )
      ,UNIQUE NONCLUSTERED ( monitorloop, blocking_spid, blocking_ecid, blocked_spid, blocked_ecid )
      )

   DECLARE @SQL NVARCHAR(MAX);
   DECLARE @TableSource NVARCHAR(MAX);

-- define source for table
   IF ( @Type = 'TABLE' ) 
      BEGIN
	-- everything input by users get quoted
         SET @TableSource = ISNULL(QUOTENAME(PARSENAME(@Trace, 4)) + N'.', '') + ISNULL(QUOTENAME(PARSENAME(@Trace, 3))
                                                                                        + N'.', '')
            + ISNULL(QUOTENAME(PARSENAME(@Trace, 2)) + N'.', '') + QUOTENAME(PARSENAME(@Trace, 1));
      END

-- define source for trc file
   IF ( @Type = 'FILE' ) 
      BEGIN	
         SET @TableSource = N'sys.fn_trace_gettable(N' + QUOTENAME(@Trace, '''') + ', -1)';
      END


-- load table or file
   IF ( @Type IN ( 'TABLE', 'FILE' ) ) 
      BEGIN
         SET @SQL = N'		
		INSERT #ReportsXML(blocked_ecid,blocked_spid,blocking_ecid,blocking_spid,
			monitorloop,bpReportXml,endTime)
		SELECT blocked_ecid,blocked_spid,blocking_ecid,blocking_spid,
			COALESCE(monitorloop, CONVERT(nvarchar(100), endTime, 120), ''unknown''),
			bpReportXml,EndTime
		FROM ' + @TableSource
            + N'
		CROSS APPLY (
			SELECT CAST(TextData as xml)
			) AS bpReports(bpReportXml)
		CROSS APPLY (
			SELECT 
				monitorloop = bpReportXml.value(''(//@monitorLoop)[1]'', ''nvarchar(100)''),
				blocked_spid = bpReportXml.value(''(/blocked-process-report/blocked-process/process/@spid)[1]'', ''int''),
				blocked_ecid = bpReportXml.value(''(/blocked-process-report/blocked-process/process/@ecid)[1]'', ''int''),
				blocking_spid = bpReportXml.value(''(/blocked-process-report/blocking-process/process/@spid)[1]'', ''int''),
				blocking_ecid = bpReportXml.value(''(/blocked-process-report/blocking-process/process/@ecid)[1]'', ''int'')
			) AS bpShredded
		WHERE EventClass = 137';
		
         EXEC (@SQL);
      END 


   IF ( @Type = 'XEVENT_RINGBUFFER' ) 
      BEGIN
		-- define source for xevent session    
         SET @TableSource = @Trace;    
		          SET @SQL = N'
         INSERT   #ReportsXML
                  ( blocked_ecid
                  ,blocked_spid
                  ,blocking_ecid
                  ,blocking_spid
                  ,monitorloop
                  ,bpReportXml
                  ,endTime
                  )
                  SELECT   n.value(''(event/data[@name="blocked_process"]/value/blocked-process-report/blocked-process/process/@ecid)[1]'',
                                   ''nvarchar(100)'') AS blocked_ecid
                          ,n.value(''(event/data[@name="blocked_process"]/value/blocked-process-report/blocked-process/process/@spid)[1]'',
                                   ''nvarchar(100)'') AS blocked_spid
                          ,n.value(''(event/data[@name="blocked_process"]/value/blocked-process-report/blocking-process/process/@ecid)[1]'',
                                   ''nvarchar(100)'') AS blocking_ecid
                          ,n.value(''(event/data[@name="blocked_process"]/value/blocked-process-report/blocking-process/process/@spid)[1]'',
                                   ''nvarchar(100)'') AS blocking_spid
                          ,n.value(''(event/data[@name="blocked_process"]/value/blocked-process-report//@monitorLoop)[1]'',
                                   ''nvarchar(100)'') AS [MonitorLoop]
                          ,n.query(''(event/data[@name="blocked_process"]/value/blocked-process-report)[1]'') AS [bpReportXml]
                          ,DATEADD(hh, DATEDIFF(hh, GETUTCDATE(), CURRENT_TIMESTAMP),
                                   n.value(''(event/@timestamp)[1]'', ''datetime'')) AS [EndTime]
                  FROM     ( SELECT td.query(''.'') AS n
                             FROM   ( SELECT CAST(target_data AS XML) AS target_data
                                      FROM   sys.dm_xe_sessions AS s
                                             JOIN sys.dm_xe_session_targets AS t ON s.address = t.event_session_address
                                      WHERE  s.name = '''+@TableSource+'''
                                             AND t.target_name = ''ring_buffer''
                                    ) AS sub
                                    CROSS APPLY target_data.nodes(''RingBufferTarget/event'') AS q ( td )
                           ) AS TAB'
          EXEC (@SQL);    
      END
 
   IF ( @Type = 'XMLFILE' ) 
      BEGIN
         CREATE TABLE #TraceXML2
            (
             id INT IDENTITY
                    PRIMARY KEY
            ,ReportXML XML NOT NULL
            )
	
         SET @SQL = N'
		INSERT #TraceXML(ReportXML)
		SELECT col FROM OPENROWSET (
				BULK ' + QUOTENAME(@Trace, '''') + N', SINGLE_BLOB
			) as xmldata(col)';

         EXEC (@SQL);
	
         CREATE PRIMARY XML INDEX PXML_TraceXML ON #TraceXML(ReportXML);

         WITH XMLNAMESPACES 
	(
		'http://tempuri.org/TracePersistence.xsd' AS MY
	),
	ShreddedWheat AS 
	(
		SELECT
			bpShredded.blocked_ecid,
			bpShredded.blocked_spid,
			bpShredded.blocking_ecid,
			bpShredded.blocking_spid,
			bpShredded.monitorloop,
			bpReports.bpReportXml,
			bpReports.bpReportEndTime
		FROM #TraceXML
		CROSS APPLY 
			ReportXML.nodes('/MY:TraceData/MY:Events/MY:Event[@name="Blocked process report"]')
			AS eventNodes(eventNode)
		CROSS APPLY 
			eventNode.nodes('./MY:Column[@name="EndTime"]')
			AS endTimeNodes(endTimeNode)
		CROSS APPLY
			eventNode.nodes('./MY:Column[@name="TextData"]')
			AS bpNodes(bpNode)
		CROSS APPLY (
			SELECT CAST(bpNode.value('(./text())[1]', 'nvarchar(max)') AS XML),
				CAST(LEFT(endTimeNode.value('(./text())[1]', 'varchar(max)'), 19) AS DATETIME)
		) AS bpReports(bpReportXml, bpReportEndTime)
		CROSS APPLY (
			SELECT 
				monitorloop = bpReportXml.value('(//@monitorLoop)[1]', 'nvarchar(100)'),
				blocked_spid = bpReportXml.value('(/blocked-process-report/blocked-process/process/@spid)[1]', 'int'),
				blocked_ecid = bpReportXml.value('(/blocked-process-report/blocked-process/process/@ecid)[1]', 'int'),
				blocking_spid = bpReportXml.value('(/blocked-process-report/blocking-process/process/@spid)[1]', 'int'),
				blocking_ecid = bpReportXml.value('(/blocked-process-report/blocking-process/process/@ecid)[1]', 'int')
		) AS bpShredded
	)
	INSERT #ReportsXML(blocked_ecid,blocked_spid,blocking_ecid,blocking_spid,
		monitorloop,bpReportXml,endTime)
	SELECT blocked_ecid,blocked_spid,blocking_ecid,blocking_spid,
		COALESCE(monitorloop, CONVERT(NVARCHAR(100), bpReportEndTime, 120), 'unknown'),
		bpReportXml,bpReportEndTime
	FROM ShreddedWheat;
	
         DROP TABLE #TraceXML

      END

-- Organize and select blocked process reports
;
   WITH  Blockheads
          AS ( SELECT   blocking_spid
                       ,blocking_ecid
                       ,monitorloop
                       ,blocking_hierarchy_string
               FROM     #ReportsXML
               EXCEPT
               SELECT   blocked_spid
                       ,blocked_ecid
                       ,monitorloop
                       ,blocked_hierarchy_string
               FROM     #ReportsXML
             ),
         Hierarchy
          AS ( SELECT   monitorloop
                       ,blocking_spid AS spid
                       ,blocking_ecid AS ecid
                       ,CAST('/' + blocking_hierarchy_string AS VARCHAR(MAX)) AS chain
                       ,0 AS level
               FROM     Blockheads
               UNION ALL
               SELECT   irx.monitorloop
                       ,irx.blocked_spid
                       ,irx.blocked_ecid
                       ,CAST(h.chain + irx.blocked_hierarchy_string AS VARCHAR(MAX))
                       ,h.level + 1
               FROM     #ReportsXML irx
                        JOIN Hierarchy h ON irx.monitorloop = h.monitorloop
                                            AND irx.blocking_spid = h.spid
                                            AND irx.blocking_ecid = h.ecid
             )
      SELECT   ISNULL(CONVERT(NVARCHAR(30), irx.endTime, 120), 'Lead') AS traceTime
              ,SPACE(4 * h.level) + CAST(h.spid AS VARCHAR(20)) + CASE h.ecid
                                                                    WHEN 0 THEN ''
                                                                    ELSE '(' + CAST(h.ecid AS VARCHAR(20)) + ')'
                                                                  END AS blockingTree
              ,irx.bpReportXml
      FROM     Hierarchy h
               LEFT JOIN #ReportsXML irx ON irx.monitorloop = h.monitorloop
                                            AND irx.blocked_spid = h.spid
                                            AND irx.blocked_ecid = h.ecid
      ORDER BY h.monitorloop
              ,h.chain

   DROP TABLE #ReportsXML

